from renats.client.subscription import Subscription
from renats.connection.base import Connection
from renats.protocol.messages.msg import MsgProtocolMessage, HMsgProtocolMessage


class EventHandler:
    def __init__(self, connection: Connection):
        self._connection: Connection = connection
        self._subscriptions: dict[str, Subscription] = {}

    def add_subscription(self, sid: str, subscription: Subscription):
        self._subscriptions[sid] = subscription

    def delete_subscription(self, sid: str):
        del self._subscriptions[sid]

    def message(self, message: MsgProtocolMessage | HMsgProtocolMessage):
        pass



import re

import msgspec.json
from msgspec import Struct, field

INFO_HEAD_PATTERN = re.compile(br"^INFO\s+(.+)\r\n$")


class InfoProtocolMessage(Struct):
    """
    NATS protocol message model for INFO message
    """
    server_id: str
    server_name: str
    version: str
    go: str
    host: str
    port: int
    headers: bool
    max_payload: int
    proto: int
    client_id: int | None = None
    auth_required: bool | None = None
    tls_required: bool | None = None
    tls_verify: bool | None = None
    tls_available: bool | None = None
    connect_urls: list[str] | None = None
    ldm: bool | None = None
    git_commit: str | None = None
    jetstream: bool | None = None
    ip: str | None = None
    client_id_str: str | None = None
    client_ip: str | None = None
    nonce: str | None = None
    cluster: str | None = None
    domain: str | None = None


class ConnectProtocolMessage(Struct, omit_defaults=True):
    """
    NATS protocol message model for CONNECT message
    """
    verbose: bool
    pedantic: bool
    tls_required: bool
    lang: str
    version: str
    auth_token: str | None = None
    user: str | None = None
    password: str | None = field(default=None, name="pass")
    name: str | None = None
    protocol: int | None = None
    echo: bool | None = None
    sig: str | None = None
    jwt: str | None = None
    no_responders: bool | None = None
    headers: bool | None = None
    nkey: str | None = None

    def dump(self) -> bytes:
        """
        Dump NATS protocol CONNECT message to bytes
        :return: NATS protocol CONNECT message as bytes-encoded string
        """
        return b"CONNECT " + msgspec.json.encode(self) + b"\r\n"



import re
from typing import Final

CRLF: Final[bytes] = b"\r\n"
CRLF_SIZE: Final[int] = len(CRLF)


def encode_headers(headers: dict[str, str]) -> dict[bytes, bytes]:
    """
    Encode headers dictionary from string-string to bytes-bytes
    :param headers: string-string headers dictionary
    :return: bytes-encoded headers dictionary
    """
    return {k.encode(): v.encode() for k, v in headers.items()}


def decode_headers(headers: dict[bytes, bytes]) -> dict[str, str]:
    """
    Decode headers dictionary from bytes-bytes to string-string
    :param headers: bytes-bytes headers dictionary
    :return: string-decoded headers dictionary
    """
    return {k.decode(): v.decode() for k, v in headers.items()}


def build_head(method: bytes, *params: bytes) -> bytes:
    """
    Build NATS protocol message head (params are joined and cleaned for multiple whitespaces)
    :param method: protocol message method (like PUB, SUB, etc...) as bytes-encoded string
    :param params: protocol message params as bytes-encoded strings
    :return: protocol message head as bytes-encoded string
    """
    return method + b" " + re.sub(br"\s{2,}", b" ", b" ".join(params))


def build_headers(headers: dict[bytes, bytes], headers_version: bytes) -> bytes:
    """
    Build NATS protocol message headers
    :param headers: dictionary with bytes-encoded headers
    :param headers_version: headers version string (NATS/1.0)
    :return: protocol message headers as bytes-encoded string
    """
    return headers_version + CRLF + CRLF.join([k + b": " + v for k, v in headers.items()])



from abc import ABC, abstractmethod
from typing import TypeVar, Callable, Any

from typing_extensions import Self

from .message import Message

HeadersType = dict[str, str]
SubscriptionCallbackType = Callable[[Message], Any]
SubscriptionType = TypeVar("SubscriptionType", bound="Subscription")


class NATS(ABC):
    @abstractmethod
    async def connect(self, servers: tuple[tuple[str, int]]) -> Self:
        raise NotImplementedError()

    @abstractmethod
    async def publish(self, subject: str, payload: bytes = b"", reply_subject: str = None, headers: HeadersType = None):
        raise NotImplementedError()

    @abstractmethod
    async def subscribe(self, subject: str, callback: SubscriptionCallbackType) -> SubscriptionType:
        raise NotImplementedError()

    @abstractmethod
    async def unsubscribe(self, subscription_id: str, messages_left: int = 0):
        raise NotImplementedError()



from msgspec import Struct

from .. import utils, protocol


class PubProtocolMessage(Struct):
    """
    NATS protocol message model for PUB message
    """
    subject: str
    reply_to: str | None = None
    payload: bytes = b""

    def dump(self) -> bytes:
        """
        Dump NATS protocol PUB message to bytes
        :return: NATS protocol PUB message as bytes-encoded string
        """
        head = utils.build_head(
            protocol.PUB,
            self.subject.encode(),
            b"" if self.reply_to is None else self.reply_to.encode(),
            str(len(self.payload)).encode()
        )
        return head + utils.CRLF + self.payload + utils.CRLF


class HPubProtocolMessage(Struct):
    """
    NATS protocol message model for HPUB message
    """
    subject: str
    headers: dict[str, str]
    reply_to: str | None = None
    payload: bytes = b""

    def dump(self) -> bytes:
        """
        Dump NATS protocol HPUB message to bytes
        :return: NATS protocol HPUB message as bytes-encoded string
        """
        headers = utils.build_headers(utils.encode_headers(self.headers), protocol.HEADERS_VERSION)
        head = utils.build_head(
            protocol.HPUB,
            self.subject.encode(),
            b"" if self.reply_to is None else self.reply_to.encode(),
            str(len(headers + utils.CRLF + utils.CRLF)).encode(),
            str(len(headers + utils.CRLF + utils.CRLF + self.payload)).encode()
        )
        return head + utils.CRLF + headers + utils.CRLF + utils.CRLF + self.payload + utils.CRLF



class NotConnected(Exception):
    def __str__(self):
        return "Not connected to server"


class UnknownProtocolMessage(Exception):
    def __init__(self, message_type: bytes):
        self.message_type = message_type

    def __str__(self):
        return f"Unknown NATS protocol message type: {self.message_type.decode()}"


class InvalidProtocolMessage(Exception):
    def __init__(self, data: bytes):
        self.data = data

    def __str__(self):
        return f"Invalid NATS protocol message: {self.data.decode()}"



class Message:
    def __init__(self, subject: str, payload: bytes, reply_subject: str = None, headers: dict[str, str] = None):
        self.subject = subject
        self.payload = payload
        self.reply_subject = reply_subject
        self.headers = headers
